/* ataxx.c: the classic Ataxx game for PALM-PILOT
 *
 * (c) 1997 Valavan Manohararajah
 */

#pragma pack(2)

#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>

/*
 * UI definitions
 */

#include "ataxx.h"





/*----------------------------Ataxx Engine-------------------------------*/

#define max_ply					4
#define max_moves				256
#define infinity				100

#if !defined(max)
#define max(a,b)				(((a) > (b)) ? (a) : (b))
#endif

#if !defined(abs)
#define abs(a)					(((a) >= 0) ? (a) : (-(a)))
#endif

#define distance(fr,fc,tr,tc)	(max(abs((fr)-(tr)),abs((fc)-(tc))))

enum {
	black,white,blank
};

enum {
	move_type_pass,move_type_grow,move_type_jump
};

typedef struct move_type {
	UInt f:7;
	UInt t:7;
	UInt type:2;
} move_t;

static Int		board[49];
static Int		turn;

static Int		num_pieces[3];

static move_t	best_mv;

static move_t	mv_lst_heap[max_ply][max_moves];
static Int		mv_score_heap[max_ply][max_moves];

static Int		save_board[max_ply][8];
static Int		save_num[max_ply];

static DWord	node_cnt;

static Int		pregen_grow_tbl[49][8];
static Int		pregen_grow_num[49];
static Int		pregen_jump_tbl[49][16];
static Int		pregen_jump_num[49];

/*
 * pregen() precomputes a set of tables used for fast move generation.
 */

#define pregen_gen(tbl,num,delta_r,delta_c)		\
	tr=r+(delta_r);								\
	tc=c+(delta_c);								\
	if(tr>=0 && tr<7 && tc>=0 && tc<7) {		\
		tbl[sq][num[sq]]=tr*7+tc;				\
		num[sq]++;								\
	}

static void pregen(void)
{
	Int r,c,tr,tc,sq=0;

	for(r=0;r<7;r++) {
		for(c=0;c<7;c++,sq++) {
			pregen_grow_num[sq]=0;
			pregen_jump_num[sq]=0;

			pregen_gen(pregen_grow_tbl,pregen_grow_num,-1, 0);
			pregen_gen(pregen_grow_tbl,pregen_grow_num,-1, 1);
			pregen_gen(pregen_grow_tbl,pregen_grow_num, 0, 1);
			pregen_gen(pregen_grow_tbl,pregen_grow_num, 1, 1);
			pregen_gen(pregen_grow_tbl,pregen_grow_num, 1, 0);
			pregen_gen(pregen_grow_tbl,pregen_grow_num, 1,-1);
			pregen_gen(pregen_grow_tbl,pregen_grow_num, 0,-1);
			pregen_gen(pregen_grow_tbl,pregen_grow_num,-1,-1);

			pregen_gen(pregen_jump_tbl,pregen_jump_num,-2, 0);
			pregen_gen(pregen_jump_tbl,pregen_jump_num,-2, 1);
			pregen_gen(pregen_jump_tbl,pregen_jump_num,-2, 2);
			pregen_gen(pregen_jump_tbl,pregen_jump_num,-1, 2);
			pregen_gen(pregen_jump_tbl,pregen_jump_num, 0, 2);
			pregen_gen(pregen_jump_tbl,pregen_jump_num, 1, 2);
			pregen_gen(pregen_jump_tbl,pregen_jump_num, 2, 2);
			pregen_gen(pregen_jump_tbl,pregen_jump_num, 2, 1);
			pregen_gen(pregen_jump_tbl,pregen_jump_num, 2, 0);
			pregen_gen(pregen_jump_tbl,pregen_jump_num, 2,-1);
			pregen_gen(pregen_jump_tbl,pregen_jump_num, 2,-2);
			pregen_gen(pregen_jump_tbl,pregen_jump_num, 1,-2);
			pregen_gen(pregen_jump_tbl,pregen_jump_num, 0,-2);
			pregen_gen(pregen_jump_tbl,pregen_jump_num,-1,-2);
			pregen_gen(pregen_jump_tbl,pregen_jump_num,-2,-2);
			pregen_gen(pregen_jump_tbl,pregen_jump_num,-2,-1);
		}
	}
}

/*
 * generate_mvs()
 */

static Int generate_mvs(Int ply)
{
	Int n,fsq,tsq,score,save_num_mvs,num_mvs=0;
	Int pregen_num,*pregen_ptr;
	move_t *mv_lst;
	Int *mv_score;

	mv_lst=mv_lst_heap[ply];
	mv_score=mv_score_heap[ply];

	for(tsq=0;tsq<49;tsq++) {
		if(board[tsq]!=blank) continue;

		score=0;
		save_num_mvs=num_mvs;

		/* grow moves */

		pregen_num=pregen_grow_num[tsq];
		pregen_ptr=pregen_grow_tbl[tsq];

		for(n=0;n<pregen_num;n++) {
			fsq=pregen_ptr[n];

			if(board[fsq]==(1-turn)) {
				score++;
			} else if((num_mvs-save_num_mvs)==0 && board[fsq]==turn) {
				mv_lst[num_mvs].f=fsq;
				mv_lst[num_mvs].t=tsq;
				mv_lst[num_mvs].type=move_type_grow;
				mv_score[num_mvs]=2;
				num_mvs++;
			}
		}

		/* jump moves */

		pregen_num=pregen_jump_num[tsq];
		pregen_ptr=pregen_jump_tbl[tsq];

		for(n=0;n<pregen_num;n++) {
			fsq=pregen_ptr[n];

			if(board[fsq]==turn) {
				mv_lst[num_mvs].f=fsq;
				mv_lst[num_mvs].t=tsq;
				mv_lst[num_mvs].type=move_type_jump;
				mv_score[num_mvs]=1;
				num_mvs++;
			}
		}

		/* score generated moves */

		for(n=save_num_mvs;n<num_mvs;n++) {
			mv_score[n]+=score;
		}
	}

	if(num_mvs==0) {
		mv_lst[0].type=move_type_pass;
		mv_score[0]=1;
		num_mvs=1;
	}

	return num_mvs;
}

/*
 * make_gains()
 */

static void make_gains(Int sq,Int ply)
{
	Int n,num=0,*ptr,pregen_num,*pregen_ptr;

	pregen_num=pregen_grow_num[sq];
	pregen_ptr=pregen_grow_tbl[sq];

	ptr=save_board[ply];

	for(n=0;n<pregen_num;n++) {
		sq=pregen_ptr[n];

		if(board[sq]==(1-turn)) {
			board[sq]=turn;
			ptr[num++]=sq;
		}
	}

	num_pieces[turn]+=num;
	num_pieces[1-turn]-=num;
	save_num[ply]=num;
}

/*
 * unmake_gains()
 */

static void unmake_gains(Int ply)
{
	Int n,num,*ptr;

	num=save_num[ply];
	ptr=save_board[ply];

	for(n=0;n<num;n++) {
		board[ptr[n]]=1-turn;
	}

	num_pieces[turn]-=num;
	num_pieces[1-turn]+=num;
}

/*
 * make_mv()
 */

static void make_mv(move_t move,Int ply)
{
	if(move.type==move_type_grow) {
		board[move.t]=turn;

		num_pieces[turn]++;
		num_pieces[blank]--;

		make_gains(move.t,ply);
	} else if(move.type==move_type_jump) {
		board[move.f]=blank;
		board[move.t]=turn;

		make_gains(move.t,ply);
	} else {
		save_num[ply]=0;
	}

	turn=1-turn;
}

/*
 * unmake_mv()
 */

static void unmake_mv(move_t move,Int ply)
{
	turn=1-turn;

	if(move.type==move_type_grow) {
		unmake_gains(ply);

		board[move.t]=blank;

		num_pieces[turn]--;
		num_pieces[blank]++;
	} else if(move.type==move_type_jump) {
		unmake_gains(ply);

		board[move.f]=turn;
		board[move.t]=blank;
	}
}

/*
 * evaluate()
 */

static Int evaluate(Int ply)
{
	if(num_pieces[blank]==0) {
		if(num_pieces[turn]>num_pieces[1-turn]) {
			return infinity-ply;
		} else {
			return -infinity+ply;
		}
	} else if(num_pieces[turn]==0) {
		return -infinity+ply;
	} else {
		return num_pieces[turn]-num_pieces[1-turn];
	}
}

/*
 * select_move()
 */

static Int select_move(Int ply,Int num,Int *mv_loc)
{
	Int score=0,loc,n,*mv_score;

	mv_score=mv_score_heap[ply];

	for(n=0;n<num;n++) {
		if(mv_score[n]>score) {
			score=mv_score[n];
			loc=n;
		}
	}

	if(score!=0) {
		mv_score[loc]=0;
		*mv_loc=loc;
		return true;
	} else {
		return false;
	}
}

/*
 * alpha_beta_i()
 */

static Int alpha_beta_i(Int alpha,Int beta,Int depth,Int ply,Int nullmove)
{
	Int score,n,num;
	move_t null_move;
	move_t *mv_lst;

	/* terminal condition */

	if(	depth<=0 || ply==max_ply ||
		num_pieces[turn]==0 || num_pieces[blank]==0) {
		return evaluate(ply);
	}

	if(nullmove) {
		null_move.type=move_type_pass;

		make_mv(null_move,ply);

		node_cnt++;

		score=-alpha_beta_i(-beta,-beta+1,depth-3,ply+1,false);

		unmake_mv(null_move,ply);

		if(score>=beta) return beta;
	}

	mv_lst=mv_lst_heap[ply];

	num=generate_mvs(ply);

	while(select_move(ply,num,&n)) {
		make_mv(mv_lst[n],ply);

		node_cnt++;

		score=-alpha_beta_i(-beta,-alpha,depth-1,ply+1,true);

		unmake_mv(mv_lst[n],ply);

		if(score>alpha) {
			if(score>=beta) return beta;
			alpha=score;
		}
	}

	return alpha;
}

/*
 * alpha_beta_r()
 */

static Int alpha_beta_r(Int depth)
{
	Int score,n,num,alpha=-infinity;
	move_t *mv_lst;
	Int *mv_score;

	mv_lst=mv_lst_heap[0];
	mv_score=mv_score_heap[0];

	num=generate_mvs(0);

	if(mv_lst[0].type==move_type_pass) {
		best_mv=mv_lst[0];
		return 0;
	}

	for(n=0;n<num;n++) {
		mv_score[n]<<=3;				/* multiply by 8 */
		mv_score[n]+=SysRandom(0) % 5;	/* randomize the move list */
	}

	while(select_move(0,num,&n)) {
		make_mv(mv_lst[n],0);

		node_cnt++;

		score=-alpha_beta_i(-infinity,-alpha,depth-1,1,true);

		unmake_mv(mv_lst[n],0);

		if(score>alpha) {
			best_mv=mv_lst[n];
			alpha=score;
		}
	}

	return alpha;
}

/*
 * board_new()
 */

static void board_new(void)
{
	Int s;

	for(s=0;s<49;s++) {
		board[s]=blank;
	}

	board[0]=board[48]=white;
	board[6]=board[42]=black;

	turn=white;

	num_pieces[blank]=45;
	num_pieces[black]=num_pieces[white]=2;
}

/*----------------------------Ataxx Engine-------------------------------*/





#define		BoardStartDrawX			0
#define		BoardStartDrawY			16

#define		BitmapStartDrawX		3
#define		BitmapStartDrawY		3

#define		SquareSizeX				16
#define		SquareSizeY				16

#define		NumberDrawWD1X			118
#define		NumberDrawWD1Y			17

#define		NumberDrawWD2X			137
#define		NumberDrawWD2Y			17

#define		NumberDrawBD1X			118
#define		NumberDrawBD1Y			114

#define		NumberDrawBD2X			137
#define		NumberDrawBD2Y			114

static Char			StatusBarString[64];

static Int			PenTrack;
static Int			PenTrackR;
static Int			PenTrackC;

static Int			SelectedF;
static Int			SelectedFR;
static Int			SelectedFC;

enum {
	game_normal,game_ended
};

enum {
	highlight_draw,highlight_erase
};

static Int			GameState;

static Int			Level;

/* System Preferences Stuff */

#define AtaxxVersion		0x0101
#define AtaxxCreator		0x71746178	/* 'Atax' */

typedef struct ProgramState_Type {
	Int		board[49];
	Int		num_pieces[3];
	Char	StatusBarString[64];
	Int		turn;
	Int		SelectedF;
	Int		SelectedFR;
	Int		SelectedFC;
	Int		GameState;
	Int		Level;
} ProgramState_t;

/*
 * NewGame()
 */

static void NewGame(void)
{
	board_new();
	SelectedF=false;
	GameState=game_normal;
	StatusBarString[0]=0;
}

/*
 * LoadState()
 */

static Boolean LoadState(void)
{
	ProgramState_t Prefs;

	if(	PrefGetAppPreferencesV10(AtaxxCreator,AtaxxVersion,&Prefs,
		sizeof(ProgramState_t))==false) {
		return false;
	}

	MemMove(board,Prefs.board,98);
	MemMove(num_pieces,Prefs.num_pieces,6);
	MemMove(StatusBarString,Prefs.StatusBarString,64);

	turn=Prefs.turn;
	SelectedF=Prefs.SelectedF;
	SelectedFR=Prefs.SelectedFR;
	SelectedFC=Prefs.SelectedFC;
	GameState=Prefs.GameState;
	Level=Prefs.Level;

	return true;
}

/*
 * SaveState()
 */

static void SaveState(void)
{
	ProgramState_t Prefs;

	MemMove(Prefs.board,board,98);
	MemMove(Prefs.num_pieces,num_pieces,6);
	MemMove(Prefs.StatusBarString,StatusBarString,64);

	Prefs.turn=turn;
	Prefs.SelectedF=SelectedF;
	Prefs.SelectedFR=SelectedFR;
	Prefs.SelectedFC=SelectedFC;
	Prefs.GameState=GameState;
	Prefs.Level=Level;

	PrefSetAppPreferencesV10(AtaxxCreator,AtaxxVersion,&Prefs,
			sizeof(ProgramState_t));
}

/*
 * StartApplication() initializes the application data
 */

static void StartApplication(void)
{
	/* prime the random number generator */

	SysRandom(TimGetSeconds());

	/* pregenerate move generation tables */

	pregen();

	/* setup for a new game at level 2 */

	if(LoadState()==false) {
		NewGame();
		Level=2;
	}
}

/*
 * BitmapDraw() is used to draw bitmaps, pass ID in.
 */

static void BitmapDraw(Word BitmapID,Int x,Int y)
{
	VoidHand	resH;
	BitmapPtr	resP;
	
	resH=DmGetResource(bitmapRsc, BitmapID);
	resP=MemHandleLock(resH);
	WinDrawBitmap(resP,x,y);
	MemPtrUnlock(resP);
	DmReleaseResource(resH);
}

/*
 * HighLight()
 */

static void HighLight(Int r,Int c,Int mode)
{
	Int x1,x2,y1,y2;

	x1=c*SquareSizeX+BoardStartDrawX+1;
	x2=x1+SquareSizeX-2;
	y1=r*SquareSizeY+BoardStartDrawY+1;
	y2=y1+SquareSizeY-2;

	if(mode==highlight_draw) {
		WinDrawLine(x1,y1,x1,y2);
		WinDrawLine(x1,y2,x2,y2);
		WinDrawLine(x2,y2,x2,y1);
		WinDrawLine(x2,y1,x1,y1);
	} else {
		WinEraseLine(x1,y1,x1,y2);
		WinEraseLine(x1,y2,x2,y2);
		WinEraseLine(x2,y2,x2,y1);
		WinEraseLine(x2,y1,x1,y1);
	}
}

/*
 * PieceDraw() is used to the pieces, it converts the row/column
 * stuff into screen co-ordinates
 */

static void PieceDraw(Word BitmapID,Int r,Int c)
{
	BitmapDraw(	BitmapID,
				c*SquareSizeX+BoardStartDrawX+BitmapStartDrawX,
				r*SquareSizeY+BoardStartDrawY+BitmapStartDrawY);
}

/*
 * BoardDraw() draws the Ataxx board.
 */

static void BoardDraw(void)
{
	BitmapDraw(Bitmap_Board,BoardStartDrawX,BoardStartDrawY);
}

/*
 * NumPiecesDraw()
 */

static void NumPiecesDraw(void)
{
	Int number;

	/* first render white number of pieces */

	number=num_pieces[white]/10;
	BitmapDraw(Bitmap_NumberW0+number,NumberDrawWD1X,NumberDrawWD1Y);
	number=num_pieces[white]%10;
	BitmapDraw(Bitmap_NumberW0+number,NumberDrawWD2X,NumberDrawWD2Y);

	/* first render white number of pieces */

	number=num_pieces[black]/10;
	BitmapDraw(Bitmap_NumberB0+number,NumberDrawBD1X,NumberDrawBD1Y);
	number=num_pieces[black]%10;
	BitmapDraw(Bitmap_NumberB0+number,NumberDrawBD2X,NumberDrawBD2Y);
}

/*
 * StatusBarDraw()
 */

static void StatusBarDraw(void)
{
	FormPtr CurrentFrm;
	FieldPtr StatusFld;

	CurrentFrm=FrmGetActiveForm();
	StatusFld=(FieldPtr)FrmGetObjectPtr(CurrentFrm,
		FrmGetObjectIndex(CurrentFrm,StatusField));

	FldSetTextPtr(StatusFld,StatusBarString);
	FldDrawField(StatusFld);
}

/*
 * MainFormDraw() redraws the main board.
 */

static void MainFormDraw(void)
{
	Int r,c,s=0;
	ControlPtr LevelPtr;
	FormPtr CurrentFrm;

	/* get a pointer to current form */

	CurrentFrm=FrmGetActiveForm();

	/* set up the level button so that it draws properly */

	switch(Level) {
	case 1:
		LevelPtr=(ControlPtr)FrmGetObjectPtr(CurrentFrm,
			FrmGetObjectIndex(CurrentFrm,Level1Button));
		break;
	case 2:
		LevelPtr=(ControlPtr)FrmGetObjectPtr(CurrentFrm,
			FrmGetObjectIndex(CurrentFrm,Level2Button));
		break;
	case 3:
		LevelPtr=(ControlPtr)FrmGetObjectPtr(CurrentFrm,
			FrmGetObjectIndex(CurrentFrm,Level3Button));
		break;
	}

	CtlSetValue(LevelPtr,1);

	/* draw the form items on the screen */

	FrmDrawForm(CurrentFrm);

	/* draw the board */

	BoardDraw();

	/* draw the status bar */

	StatusBarDraw();

	/* draw the piece counts */

	NumPiecesDraw();

	/* draw the pieces */

	for(r=0;r<7;r++) {
		for(c=0;c<7;c++,s++) {
			switch(board[s]) {
			case black:
				PieceDraw(Bitmap_BPieceN,r,c);
				break;
			case white:
				PieceDraw(Bitmap_WPieceN,r,c);
				break;
			}
		}
	}

	/* if a piece is currently selected, highlight it */

	if(SelectedF==true) {
		HighLight(SelectedFR,SelectedFC,highlight_draw);
	}
}

/*
 * WhiteToBlackAnimate()
 */

static void WhiteToBlackAnimate(Int *change_board,Int change_num)
{
	Int n;

	for(n=0;n<change_num;n++) {
		PieceDraw(Bitmap_WPieceA,change_board[n]/7,change_board[n]%7);
	}

	SysTaskDelay(12);

	for(n=0;n<change_num;n++) {
		PieceDraw(Bitmap_PieceT,change_board[n]/7,change_board[n]%7);
	}

	SysTaskDelay(12);

	for(n=0;n<change_num;n++) {
		PieceDraw(Bitmap_BPieceA,change_board[n]/7,change_board[n]%7);
	}

	SysTaskDelay(12);
	
	for(n=0;n<change_num;n++) {
		PieceDraw(Bitmap_BPieceN,change_board[n]/7,change_board[n]%7);
	}
}

/*
 * BlackToWhiteAnimate()
 */

static void BlackToWhiteAnimate(Int *change_board,Int change_num)
{
	Int n;

	for(n=0;n<change_num;n++) {
		PieceDraw(Bitmap_BPieceA,change_board[n]/7,change_board[n]%7);
	}

	SysTaskDelay(12);

	for(n=0;n<change_num;n++) {
		PieceDraw(Bitmap_PieceT,change_board[n]/7,change_board[n]%7);
	}

	SysTaskDelay(12);

	for(n=0;n<change_num;n++) {
		PieceDraw(Bitmap_WPieceA,change_board[n]/7,change_board[n]%7);
	}

	SysTaskDelay(12);
	
	for(n=0;n<change_num;n++) {
		PieceDraw(Bitmap_WPieceN,change_board[n]/7,change_board[n]%7);
	}
}

/*
 * GameStateUpdate()
 */

static void GameStateUpdate(void)
{
	Int eval;

	eval=evaluate(0);

	if(abs(eval)==infinity) {
		GameState=game_ended;
		if(	(turn==white && eval==infinity) ||
			(turn==black && eval==-infinity)) {
			StrCopy(StatusBarString,"White Wins");
		} else {
			StrCopy(StatusBarString,"Black Wins");
		}
	} else {
		GameState=game_normal;
		StatusBarString[0]=0;
	}

	/* draw the status bar */

	StatusBarDraw();

	/* draw the piece counts */

	NumPiecesDraw();
}

/*
 * ComputerMoveMake()
 */

static void ComputerMoveMake()
{
	Int fr,fc,tr,tc;

	if(GameState==game_normal) {
		alpha_beta_r(Level);

		make_mv(best_mv,0);

		if(best_mv.type==move_type_pass) {
			StrCopy(StatusBarString,"Computer Passes");
			StatusBarDraw();
		} else {
			fr=best_mv.f/7;
			fc=best_mv.f%7;

			tr=best_mv.t/7;
			tc=best_mv.t%7;

			HighLight(fr,fc,highlight_draw);
			SysTaskDelay(30);

			HighLight(tr,tc,highlight_draw);
			SysTaskDelay(30);

			if(best_mv.type==move_type_jump) {
				PieceDraw(Bitmap_PieceE,fr,fc);
			}

			HighLight(fr,fc,highlight_erase);
			HighLight(tr,tc,highlight_erase);

			switch(board[best_mv.t]) {
			case black:
				PieceDraw(Bitmap_BPieceN,tr,tc);
				if(save_num[0]) {
					WhiteToBlackAnimate(save_board[0],save_num[0]);
				}
				break;
			case white:
				PieceDraw(Bitmap_WPieceN,tr,tc);
				if(save_num[0]) {
					BlackToWhiteAnimate(save_board[0],save_num[0]);
				}
				break;
			}

			GameStateUpdate();
		}
	}
}

/*
 * UserPassMake()
 */

static void UserPassMake(void)
{
	Int num;

	if(GameState==game_normal) {
		num=generate_mvs(0);
		if(mv_lst_heap[0][0].type==move_type_pass) {
			if(SelectedF==true) {
				HighLight(SelectedFR,SelectedFC,highlight_erase);
			}

			make_mv(mv_lst_heap[0][0],0);
			GameStateUpdate();

			ComputerMoveMake();

			return;
		}
	}

	/* display illegal to pass dialog */

	FrmAlert(PassAlert);
}

/*
 * UserMoveMake()
 */

static void UserMoveMake(Int fr,Int fc,Int tr,Int tc,Int dist)
{
	move_t move;

	move.f=fr*7+fc;
	move.t=tr*7+tc;

	if(dist==2) {
		move.type=move_type_jump;
		PieceDraw(Bitmap_PieceE,fr,fc);
	} else {
		move.type=move_type_grow;
	}

	make_mv(move,0);

	switch(board[move.t]) {
	case black:
		PieceDraw(Bitmap_BPieceN,tr,tc);
		if(save_num[0]) {
			WhiteToBlackAnimate(save_board[0],save_num[0]);
		}
		break;
	case white:
		PieceDraw(Bitmap_WPieceN,tr,tc);
		if(save_num[0]) {
			BlackToWhiteAnimate(save_board[0],save_num[0]);
		}
		break;
	}

	HighLight(SelectedFR,SelectedFC,highlight_erase);

	GameStateUpdate();

	ComputerMoveMake();
}

#define PenTrackUpdate() \
	PenTrackR=(event->screenY-BoardStartDrawY)/SquareSizeY; \
	PenTrackC=(event->screenX-BoardStartDrawX)/SquareSizeX;

#define PenTrackValid() \
	(PenTrackR>=0 && PenTrackR<7 && PenTrackC>=0 && PenTrackC<7)

/*
 * MainFormPenDownHandler()
 */

static Boolean MainFormPenDownHandler(EventPtr event)
{
	PenTrackUpdate();

	if(GameState==game_normal && PenTrackValid()) {
		/* valid square has been received, start tracking */

		PenTrack=true;
		HighLight(PenTrackR,PenTrackC,highlight_draw);

		return true;
	} else {
		/* invalid square has been received, disable tracking */

		PenTrack=false;

		return false;
	}
}

/*
 * MainFormPenMoveHandler()
 */

static Boolean MainFormPenMoveHandler(EventPtr event)
{
	if(PenTrack==true) {
		if(	PenTrackValid() &&
			(!SelectedF || SelectedFR!=PenTrackR ||
			 SelectedFC!=PenTrackC)) {
			HighLight(PenTrackR,PenTrackC,highlight_erase);
		}

		PenTrackUpdate();

		if(PenTrackValid()) {
			HighLight(PenTrackR,PenTrackC,highlight_draw);
		}

		return true;
	} else {
		return false;
	}
}

/*
 * MainFormPenUpHandler()
 */

static Boolean MainFormPenUpHandler(EventPtr event)
{
	Int dist;

	if(PenTrack==true) {
		if(PenTrackValid()) {
			HighLight(PenTrackR,PenTrackC,highlight_erase);
		}

		PenTrackUpdate();

		if(PenTrackValid()) {
			if(SelectedF==true) {
				/* valid first square selection exists */

				if(board[PenTrackR*7+PenTrackC]==turn) {
					/* user wants to switch to another piece */

					HighLight(SelectedFR,SelectedFC,highlight_erase);

					SelectedFR=PenTrackR;
					SelectedFC=PenTrackC;

					HighLight(SelectedFR,SelectedFC,highlight_draw);
				} else {
					if(board[PenTrackR*7+PenTrackC]==blank) {
						dist=distance(	PenTrackR,PenTrackC,
										SelectedFR,SelectedFC);

						if(dist<=2) {
							/* valid */
							UserMoveMake(	SelectedFR,SelectedFC,
											PenTrackR,PenTrackC,dist);

							SelectedF=false;
						}
					}
				}
			} else {
				/* user is making "from" square selection */

				if(board[PenTrackR*7+PenTrackC]==turn) {
					/* user makes valid from square selection */

					SelectedF=true;
					SelectedFR=PenTrackR;
					SelectedFC=PenTrackC;

					HighLight(SelectedFR,SelectedFC,highlight_draw);
				}
			}
		}
		return true;
	} else {
		return false;
	}
}

/*
 * MainFormHandler() handles all events associated with the main form.
 */

static Boolean MainFormHandler(EventPtr event)
{
	switch(event->eType) {
	case frmOpenEvent:
		MainFormDraw();
		return true;
	case frmUpdateEvent:
		MainFormDraw();
		return true;
	case ctlSelectEvent:
		switch(event->data.ctlSelect.controlID) {
		case NewButton:		
			NewGame();
			MainFormDraw();
			break;
		case SwitchButton:
			if(SelectedF==true) {
				HighLight(SelectedFR,SelectedFC,highlight_erase);
			}
			ComputerMoveMake();
			break;
		case PassButton:
			UserPassMake();
			break;
		case HelpButton:
			FrmHelp(HelpStr);
			break;
		case AboutButton:
			FrmGotoForm(AboutForm);
			break;
		case Level1Button:
			Level=1;
			break;
		case Level2Button:
			Level=2;
			break;
		case Level3Button:
			Level=3;
			break;
		}
		return true;
	case penDownEvent:
		return MainFormPenDownHandler(event);
	case penMoveEvent:
		return MainFormPenMoveHandler(event);
	case penUpEvent:
		return MainFormPenUpHandler(event);
	default:
		return false;
	}
}

/*
 * AboutFormHandler() handles all events associated with the about form.
 */

static Boolean AboutFormHandler(EventPtr event)
{
	switch(event->eType) {
	case frmOpenEvent:
	case frmUpdateEvent:
		FrmDrawForm(FrmGetActiveForm());
		return true;
	case ctlSelectEvent:
		switch(event->data.ctlSelect.controlID) {
		case AboutOkButton:
			FrmGotoForm(MainForm);
			break;
		}
		return true;
	default:
		return false;
	}
}

/*
 * ApplicationHandleEvent() handles any application related events.
 */

static Boolean ApplicationHandleEvent(EventPtr event)
{
	FormPtr CurrentFrm;

	if(event->eType==frmLoadEvent) {
		CurrentFrm=FrmInitForm(event->data.frmLoad.formID);
		FrmSetActiveForm(CurrentFrm);

		switch(event->data.frmLoad.formID) {
		case MainForm:
			FrmSetEventHandler(CurrentFrm,MainFormHandler);
			break;
		case AboutForm:
			FrmSetEventHandler(CurrentFrm,AboutFormHandler);
			break;
		}

		return true;
	} else {
		return false;
	}
}

/*
 * EventLoop() fetchs new events from the event queue and dispatches them.
 */

static void EventLoop(void)
{
	Word error;
	EventType event;

	do {
		EvtGetEvent(&event,evtWaitForever);
		if(!SysHandleEvent(&event)) {
			if(!MenuHandleEvent((void *)0,&event,&error)) {
				if(!ApplicationHandleEvent(&event)) {
					FrmDispatchEvent(&event);
				}
			}
		}
	} while(event.eType!=appStopEvent);
}

/*
 * StopApplication() houses all clean up code
 */

static void StopApplication(void)
{
	SaveState();
}

/*
 * PILOT entry point
 */ 

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
	/* if not a normal launch then return to OS */

	if(cmd) return 0;

	/* initialize the app */

	StartApplication();

	/* show the main form */

    FrmGotoForm(MainForm);

	/* event processor */

	EventLoop();
	
	/* cleanup code */

	StopApplication ();

	return 0;
}